@@ -20,7 +20,9 @@ from octavia_lib.api.drivers import exceptions as driver_exceptions
 from octavia_lib.api.drivers import provider_base as driver_base
 from octavia_lib.common import constants
 from oslo_log import log as logging
+from ovsdbapp.backend.ovs_idl import idlutils
 
+from ovn_octavia_provider.common import clients
 from ovn_octavia_provider.common import config as ovn_conf
 # TODO(mjozefcz): Start consuming const and utils
 # from neutron-lib once released.
@@ -88,7 +90,7 @@ class OvnProviderDriver(driver_base.ProviderDriver):
                 user_fault_string=msg,
                 operator_fault_string=msg)
 
-    def loadbalancer_create(self, loadbalancer):
+    def _get_loadbalancer_request_info(self, loadbalancer):
         admin_state_up = loadbalancer.admin_state_up
         if isinstance(admin_state_up, o_datamodels.UnsetType):
             admin_state_up = True
@@ -101,9 +103,97 @@ class OvnProviderDriver(driver_base.ProviderDriver):
                           o_datamodels.UnsetType):
             request_info[constants.ADDITIONAL_VIPS] = \
                 loadbalancer.additional_vips
+        return request_info
 
+    def _get_listener_request_info(self, listener):
+        self._check_for_supported_protocols(listener.protocol)
+        self._check_for_allowed_cidrs(listener.allowed_cidrs)
+        admin_state_up = listener.admin_state_up
+        if isinstance(admin_state_up, o_datamodels.UnsetType):
+            admin_state_up = True
+        request_info = {'id': listener.listener_id,
+                        'protocol': listener.protocol,
+                        'loadbalancer_id': listener.loadbalancer_id,
+                        'protocol_port': listener.protocol_port,
+                        'default_pool_id': listener.default_pool_id,
+                        'admin_state_up': admin_state_up}
+        return request_info
+
+    def _get_pool_request_info(self, pool):
+        self._check_for_supported_protocols(pool.protocol)
+        self._check_for_supported_algorithms(pool.lb_algorithm)
+        admin_state_up = pool.admin_state_up
+        if isinstance(admin_state_up, o_datamodels.UnsetType):
+            admin_state_up = True
+        request_info = {'id': pool.pool_id,
+                        'loadbalancer_id': pool.loadbalancer_id,
+                        'protocol': pool.protocol,
+                        'lb_algorithm': pool.lb_algorithm,
+                        'listener_id': pool.listener_id,
+                        'admin_state_up': admin_state_up}
+        if not isinstance(
+                pool.session_persistence, o_datamodels.UnsetType):
+            self._check_for_supported_session_persistence(
+                pool.session_persistence)
+            request_info['session_persistence'] = pool.session_persistence
+
+        return request_info
+
+    def _get_member_request_info(self, member, create=True):
+        # Validate monitoring options if present
+        admin_state_up = None
+        if create:
+            self._check_member_monitor_options(member)
+            if self._ip_version_differs(member):
+                raise ovn_exc.IPVersionsMixingNotSupportedError()
+            admin_state_up = member.admin_state_up
+        subnet_id = member.subnet_id
+        if (isinstance(subnet_id, o_datamodels.UnsetType) or not subnet_id):
+            subnet_id, subnet_cidr = self._ovn_helper._get_subnet_from_pool(
+                member.pool_id)
+            if not (subnet_id and
+                    self._ovn_helper._check_ip_in_subnet(member.address,
+                                                         subnet_cidr)):
+                msg = _('Subnet is required, or Loadbalancer associated with '
+                        'Pool must have a subnet, for Member creation '
+                        'with OVN Provider Driver if it is not the same as '
+                        'LB VIP subnet')
+                raise driver_exceptions.UnsupportedOptionError(
+                    user_fault_string=msg,
+                    operator_fault_string=msg)
+
+        if isinstance(admin_state_up, o_datamodels.UnsetType):
+            admin_state_up = True
+        request_info = {'id': member.member_id,
+                        'address': member.address,
+                        'protocol_port': member.protocol_port,
+                        'pool_id': member.pool_id,
+                        'subnet_id': subnet_id}
+
+        if admin_state_up and create:
+            request_info['admin_state_up'] = admin_state_up
+
+        return request_info
+
+    def _get_healthmonitor_request_info(self, healthmonitor):
+        self._validate_hm_support(healthmonitor)
+        admin_state_up = healthmonitor.admin_state_up
+        if isinstance(admin_state_up, o_datamodels.UnsetType):
+            admin_state_up = True
+        request_info = {'id': healthmonitor.healthmonitor_id,
+                        'pool_id': healthmonitor.pool_id,
+                        'type': healthmonitor.type,
+                        'interval': healthmonitor.delay,
+                        'timeout': healthmonitor.timeout,
+                        'failure_count': healthmonitor.max_retries_down,
+                        'success_count': healthmonitor.max_retries,
+                        'admin_state_up': admin_state_up}
+        return request_info
+
+    def loadbalancer_create(self, loadbalancer):
         request = {'type': ovn_const.REQ_TYPE_LB_CREATE,
-                   'info': request_info}
+                   'info': self._get_loadbalancer_request_info(
+                       loadbalancer)}
         self._ovn_helper.add_request(request)
 
         if not isinstance(loadbalancer.listeners, o_datamodels.UnsetType):
@@ -628,3 +718,227 @@ class OvnProviderDriver(driver_base.ProviderDriver):
         request = {'type': ovn_const.REQ_TYPE_HM_DELETE,
                    'info': request_info}
         self._ovn_helper.add_request(request)
+
+    def _ensure_loadbalancer(self, loadbalancer):
+        try:
+            ovn_lbs = self._ovn_helper._find_ovn_lbs_with_retry(
+                loadbalancer.loadbalancer_id)
+        except idlutils.RowNotFound:
+            LOG.debug(f"OVN loadbalancer {loadbalancer.loadbalancer_id} "
+                      "not found. Start create process.")
+            # TODO(froyo): By now just syncing LB, listener and pool only
+            status = self._ovn_helper.lb_create(
+                self._get_loadbalancer_request_info(loadbalancer))
+
+            if not isinstance(loadbalancer.listeners, o_datamodels.UnsetType):
+                status[constants.LISTENERS] = []
+                for listener in loadbalancer.listeners:
+                    status_listener = self._ovn_helper.listener_create(
+                        self._get_listener_request_info(listener))
+                    status[constants.LISTENERS].append(
+                        status_listener[constants.LISTENERS][0])
+            if not isinstance(loadbalancer.pools, o_datamodels.UnsetType):
+                status[constants.POOLS] = []
+                for pool in loadbalancer.pools:
+                    status_pool = self._ovn_helper.pool_create(
+                        self._get_pool_request_info(pool))
+                    status[constants.POOLS].append(
+                        status_pool[constants.POOLS][0])
+                    for member in pool.members:
+                        status[constants.MEMBERS] = []
+                        if not member.subnet_id:
+                            member.subnet_id = loadbalancer.vip_subnet_id
+                        status_member = self._ovn_helper.member_create(
+                            [self._get_member_request_info(member)])
+                        status[constants.MEMBERS].append(
+                            status_member[constants.MEMBERS][0])
+                    if pool.healthmonitor is not None and not isinstance(
+                            pool.healthmonitor, o_datamodels.UnsetType):
+                        status[constants.HEALTHMONITORS] = []
+                        lbhcs, ovn_hm_lb = (
+                            self._ovn_helper._find_ovn_lb_from_hm_id(
+                                pool.healthmonitor.healthmonitor_id)
+                        )
+                        if not lbhcs and ovn_hm_lb is None:
+                            status_hm = self._ovn_helper.hm_create(
+                                self._get_healthmonitor_request_info(
+                                    pool.healthmonitor))
+                            status[constants.HEALTHMONITORS].append(
+                                status_hm[constants.HEALTHMONITORS][0])
+            self._ovn_helper._update_status_to_octavia(status)
+        else:
+            # Load Balancer found, check LB and listener/pool/member/hms
+            # related
+            for ovn_lb in ovn_lbs:
+                LOG.debug(
+                    f"Sync - Loadbalancer {loadbalancer.loadbalancer_id} "
+                    "found checking other entities related")
+                self._ovn_helper.lb_sync(
+                    self._get_loadbalancer_request_info(loadbalancer), ovn_lb)
+                # Listener
+                if not isinstance(loadbalancer.listeners,
+                                  o_datamodels.UnsetType):
+                    for listener in loadbalancer.listeners:
+                        self._ovn_helper.listener_sync(
+                            self._get_listener_request_info(listener), ovn_lb)
+                # Pool
+                if not isinstance(loadbalancer.pools, o_datamodels.UnsetType):
+                    for pool in loadbalancer.pools:
+                        pool_info = self._get_pool_request_info(pool)
+                        self._ovn_helper.pool_sync(pool_info, ovn_lb)
+                        ovn_pool_key = self._ovn_helper._get_pool_key(
+                            pool_info[constants.ID],
+                            is_enabled=pool_info[constants.ADMIN_STATE_UP])
+                        member_ids = []
+                        if not isinstance(pool.members,
+                                          o_datamodels.UnsetType):
+                            for member in pool.members:
+                                if not member.subnet_id:
+                                    member.subnet_id = (
+                                        loadbalancer.vip_subnet_id
+                                    )
+                                self._ovn_helper.member_sync(
+                                    self._get_member_request_info(member),
+                                    ovn_lb,
+                                    ovn_pool_key)
+                                member_ids.append(member.member_id)
+
+                            for ovn_mb_info in \
+                                self._ovn_helper._get_members_in_ovn_lb(
+                                    ovn_lb, ovn_pool_key):
+                                # If member ID not in pool member list,
+                                # delete it.
+                                if ovn_mb_info[3] not in member_ids:
+                                    LOG.debug(
+                                        "Start deleting extra member "
+                                        f"{ovn_mb_info[3]} from pool "
+                                        "{pool_info[constants.ID]} in OVN."
+                                    )
+                                    mb_delete_info = {
+                                        'id': ovn_mb_info[3],
+                                        'subnet_id': ovn_mb_info[2],
+                                    }
+                                    self._ovn_helper.member_delete(
+                                        [mb_delete_info])
+
+                                    mb_delete_dvr_info = {
+                                        'id': ovn_mb_info[3],
+                                        'address': ovn_mb_info[0],
+                                        'pool_id': pool_info[constants.ID],
+                                        'subnet_id': ovn_mb_info[2],
+                                        'action':
+                                            ovn_const.REQ_INFO_MEMBER_DELETED
+                                    }
+                                    self._ovn_helper.handle_member_dvr(
+                                        [mb_delete_dvr_info])
+                        # Check health monitor
+                        if pool.healthmonitor is not None and not isinstance(
+                                pool.healthmonitor, o_datamodels.UnsetType):
+                            self._ovn_helper.hm_sync(
+                                self._get_healthmonitor_request_info(
+                                    pool.healthmonitor),
+                                ovn_lb,
+                                ovn_pool_key)
+                # Purge HM
+                self._ovn_helper.hm_purge(loadbalancer.loadbalancer_id)
+                status = self._ovn_helper._get_current_operating_statuses(
+                    ovn_lb)
+                self._ovn_helper._update_status_to_octavia(status)
+
+    def _fip_sync(self, loadbalancer):
+        LOG.info("Starting sync floating IP for loadbalancer "
+                 f"{loadbalancer.loadbalancer_id}")
+        if not loadbalancer.vip_port_id or not loadbalancer.vip_network_id:
+            LOG.debug("VIP Port or Network not set for loadbalancer "
+                      f"{loadbalancer.loadbalancer_id}, skip FIP sync.")
+            return
+
+        # Try to get FIP from neutron
+        fips = self._ovn_helper.get_fip_from_vip(loadbalancer)
+
+        # get FIP from LSP
+        vip_lsp = self._ovn_helper.get_lsp(
+            port_id=loadbalancer.vip_port_id,
+            network_id=loadbalancer.vip_network_id)
+        lsp_fip = vip_lsp.external_ids.get(
+            ovn_const.OVN_PORT_FIP_EXT_ID_KEY) if vip_lsp else None
+
+        if fips:
+            neutron_fip = fips[0].floating_ip_address
+            if not vip_lsp:
+                LOG.warn(
+                    "Logic Switch Port not found for port "
+                    f"{loadbalancer.vip_port_id}. "
+                    "Skip sync FIP for loadbalancer "
+                    f"{loadbalancer.loadbalancer_id}. Please "
+                    "run command `neutron-ovn-db-sync-util` "
+                    "first to sync OVN DB with Neutron DB.")
+                return
+            if lsp_fip != neutron_fip:
+                LOG.warn(
+                    "Floating IP not consistent between Logic Switch "
+                    f"Port and Neutron. Found FIP {lsp_fip} "
+                    f"in LSP {vip_lsp.name}, but we have {neutron_fip} from "
+                    "Neutron. Skip sync FIP for "
+                    f"loadbalancer {loadbalancer.loadbalancer_id}. "
+                    "Please run command `neutron-ovn-db-sync-util` "
+                    "first to sync OVN DB with Neutron DB.")
+                return
+            self._ovn_helper.vip_port_update_handler(
+                vip_lp=vip_lsp, fip=lsp_fip,
+                action=ovn_const.REQ_INFO_ACTION_SYNC)
+        else:
+            LOG.warn("Floating IP not found for loadbalancer "
+                     f"{loadbalancer.loadbalancer_id}")
+            if lsp_fip:
+                LOG.warn(
+                    "Floating IP not consistent between Logic Switch "
+                    f"Port and Neutron. Found FIP {lsp_fip} configured "
+                    f"in LSP {vip_lsp.name}, but no FIP configured from "
+                    "Neutron. Please run command `neutron-ovn-db-sync-util` "
+                    "first to sync OVN DB with Neutron DB.")
+
+    def do_sync(self, **lb_filters):
+        LOG.info(f"Starting sync OVN DB with Loadbalancer filter {lb_filters}")
+        octavia_client = clients.get_octavia_client()
+        # We can add project_id to lb_filters for lbs to limit the scope.
+        lbs = self._ovn_helper.get_octavia_lbs(octavia_client, **lb_filters)
+        for lb in lbs:
+            LOG.info(f"Starting sync OVN DB with Loadbalancer {lb.name}")
+            provider_lb = (
+                self._ovn_helper._octavia_driver_lib.get_loadbalancer(lb.id)
+            )
+
+            listeners = provider_lb.listeners or []
+            provider_lb.listeners = [
+                o_datamodels.Listener.from_dict(listener)
+                for listener in listeners
+            ] if listeners else o_datamodels.Unset
+
+            pools = provider_lb.pools or []
+            provider_pools = []
+            for pool in pools:
+                provider_pool = o_datamodels.Pool.from_dict(pool)
+                # format member provider
+                members = provider_pool.members
+                if not isinstance(members, o_datamodels.UnsetType) and members:
+                    provider_pool.members = [
+                        o_datamodels.Member.from_dict(m)
+                        for m in members]
+                else:
+                    provider_pool.members = o_datamodels.Unset
+
+                # format healthmonitor provider
+                if not isinstance(
+                        provider_pool.healthmonitor, o_datamodels.UnsetType
+                ) and provider_pool.healthmonitor is not None:
+                    provider_pool.healthmonitor = \
+                        o_datamodels.HealthMonitor.from_dict(
+                            provider_pool.healthmonitor)
+                provider_pools.append(provider_pool)
+
+            provider_lb.pools = (
+                provider_pools if provider_pools else o_datamodels.Unset
+            )
+            self._ensure_loadbalancer(provider_lb)
+            self._fip_sync(provider_lb)
